import java.net.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;
import java.awt.event.*;
import java.awt.Dimension;

public class Cliente extends JFrame implements ActionListener, Runnable{
    // Componentes interfase usuario
    private JTextArea txtCharla;    // Area de texto con todo el chat
    private JScrollPane scrCharla;  // Panel para contener el area texto con scrollBar
    private DefaultListModel lm;    // Clase que implementa lista sobre vector 
                    //  dinamico y notifica a los escuchas cambios en la lista
    private JList lstConectados;    // Lista de usuarios conectados, selec mltiple
    private JScrollPane scrConect;  // Lista desplegable
    private JTextField txtEnvio;    // Texto a enviar
    private JButton btnEnviar;      // Boton para evento enviar
    private JSplitPane splitPane;   // Panel dividido p/2 componentes, dimensionable
    private JToolBar tbBarra;       // barra de herramientas
    private JButton btnConectar;    // Boton para pedir ventana p/conexion con servidor
    private JButton btnCerrar;      // Boton para cerrar conexin
    private JLabel lblUsuarios;     // Etiquetas de usuarios
    
    // Objetos
    private String mensaje; // Texto a enviar a todos los clientes
    private String nick;    // Apodo, alias del usuario cliente
    private Socket servidor = null; // Puerto de escucha en el servidor
    private DataInputStream entrada = null; // Flujo de entrada (servidor al cliente) 
    private DataOutputStream salida = null; // Flujo de salida (cliente al servidor)
    private Thread receptor;    // Objeto receptor
    private boolean conectado = false;
		

    public Cliente(){   // Constructor
	btnConectar = new JButton("Conectar...",new ImageIcon("conectar.gif"));
        btnConectar.addActionListener(this);
	lblUsuarios = new JLabel("Desconectado", new ImageIcon("usuarios.gif"),JLabel.LEFT);
	lblUsuarios.setEnabled(false);
        
        // Barra de tareas
        
	tbBarra = new JToolBar();   // Instanciamos 
        tbBarra.setBounds(0,0,600,45);  //  Dimensionamos
	tbBarra.setFloatable(false);    //  No podr ser arrastrada, movida
	tbBarra.add(btnConectar);     // Agregamos boton conectar a la barra
	tbBarra.addSeparator();       // Separamos      
	JButton botonTB = new JButton(new ImageIcon("limpiar.gif"));
	botonTB.setToolTipText("Limpiar rea de conversacin");// Microtexto
	botonTB.addActionListener(new ActionListener(){ // Respuesta al evento
            public void actionPerformed(ActionEvent e){txtCharla.setText("");}
	}); 
	tbBarra.add(botonTB); // Lo incorporamos a la barra
	tbBarra.addSeparator(); //  Separacion
	botonTB = new JButton(new ImageIcon("acercaDe.gif"));
	botonTB.setToolTipText("Acerca de");
	botonTB.addActionListener(new ActionListener(){ //  Respuesta a este evento
            public void actionPerformed(ActionEvent e){acercaDe();}
	});
	tbBarra.add(botonTB);   // Incorporamos a la barra
	btnCerrar = new JButton(new ImageIcon("cerrar.gif"));
	btnCerrar.setToolTipText("Salir");
	btnCerrar.addActionListener(new ActionListener(){//  Respuesta a este evento
            public void actionPerformed(ActionEvent e){
		if(conectado){conectado=!desconectar();}
                    dispose();  //  Metodo heredado de java.awt.Windows. Libera
                                //  todos los recursos de este componente
		System.exit(0);
            }
	});
	tbBarra.add(btnCerrar);
	tbBarra.addSeparator(new Dimension(150,10));
	tbBarra.add(lblUsuarios);   //  Etiqueta para mensajes a usuarios
        
        //  Area de conversacin
        
	txtCharla = new JTextArea();    //  Instanciamos
	txtCharla.setEditable(false);   //  No editable (Solo salida)
	txtCharla.setLineWrap(true);    //  Desplazamiento vertical
	txtCharla.setWrapStyleWord(true);// Envoltura usado en lineas desplazables
	txtCharla.setToolTipText("Conversacin en curso");
	scrCharla = new JScrollPane(txtCharla); // Incorporamos en contenedor
	scrCharla.setMinimumSize(new Dimension(100,50));// definimos tamao mnimo
        
        //  Lista de usuarios (Conectados)
        
	lm = new DefaultListModel();    // Definimos modelo de lista, con el
	lstConectados = new JList(lm);    // instanciamos lista de usuarios  
	lstConectados.setSelectionMode(0);  // Seleccion simple
	lstConectados.setToolTipText("Usuarios conectados");
        
        //  Mecanica de envio de mensajes privados entre los usuarios conectados
        
	lstConectados.addListSelectionListener(new ListSelectionListener(){
            /*  Incorporamos y desarrollamos el escucha. El metodo valueChanged
             *  responde al evento cambio de seleccin en JList             */
            public void valueChanged(ListSelectionEvent e){
		String nickdest=lstConectados.getSelectedValue().toString();
                // A quien de los usuarios seleccionamos (para msge privado) 
		if(nickdest!=null && !nickdest.startsWith(">> ")){
                    //  Si no me seleccione (a mi mismo), ventana para dialogo
                    String msg = JOptionPane.showInputDialog("Mensaje privado para "+nickdest);
                    if(msg!=null && !msg.matches("")){ //   si algo escribo
                        //  preparo el JTextField y lo envio a todos sus escuchas 
			txtEnvio.setText("/p"+lstConectados.getSelectedValue()+"\t"+msg);
 			txtEnvio.postActionEvent();
                    }
		}
            }
	});
	scrConect = new JScrollPane(lstConectados); // incorporamos al contenedor
	scrConect.setMinimumSize(new Dimension(50,50));
        
        /* Usaremos un objeto JSplitPane, (Panel dividido) que permite situar 
         * dos componentes, el izquierdo el area de conversacin y el derecho
         * la lista de usuarios conectados. Dentro de el, podemos redimensionar 
         * el tamao de los componentes dinamicamente                       */
        
	splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT); // Distrib. horiz.
    	splitPane.setLeftComponent(scrCharla);  //  Situamos comp. izquierdo
    	splitPane.setRightComponent(scrConect); //  Idem derecho
    	splitPane.setOneTouchExpandable(true);  //  Usuario puede redimensionar
        splitPane.setDividerLocation(490);      //  Posicion de la divisoria
    	splitPane.setDividerSize(10);           //  Ancho de la divisoria
     	splitPane.setBounds(2,45,600,360);      //  Dimensiones 
        
        //  El campo de texto p/mensaje y boton enviar
        
	txtEnvio = new JTextField();
	txtEnvio.setBounds(2,410,490,25);
	txtEnvio.addActionListener(this);
	txtEnvio.setToolTipText("Escriba aqu su mensaje");
	btnEnviar = new JButton("Enviar");
	btnEnviar.setEnabled(false);
	btnEnviar.setBounds(497,410,105,25);
	btnEnviar.setToolTipText("Presione aqu para enviar su mensaje");
	btnEnviar.addActionListener(this);
        
        // Incorporamos todos los componentes a la ventana principal
        
	getContentPane().setLayout(null);
	getContentPane().add(splitPane);
	getContentPane().add(txtEnvio);
	getContentPane().add(btnEnviar);
	getContentPane().add(tbBarra);
	this.addWindowListener(new CierroVentana()); //  Escucha de la ventana
	txtEnvio.requestFocus();
    }   //  constructor (public Cliente)   
	
    public void acercaDe(){
	JOptionPane.showMessageDialog(this, "RS Chat v1.0\n\nDesarrollado por:\n<< Rodrigo Nogus >>\npara el final de PPR.","Acerca de RS Chat",JOptionPane.INFORMATION_MESSAGE,new ImageIcon("usuarios.gif"));
    }

    /* A continuacin definimos la respuesta a eventos a nivel ventana cliente 
     * (JFrame), o sea todos aquellos que no tienen su escucha definida junto 
     * al componente botonTB, btnCerrar, por ejemplo)                       */  
    
    public void actionPerformed(ActionEvent evt){
	if(evt.getActionCommand().equals("Desconectar")){
            conectado=!desconectar();
            return;
	}
		
	if(conectado){
            String texto = txtEnvio.getText();
            texto.trim();
            if (texto.length()>0){
		try{
                    salida.writeUTF(texto); //  Enviamos al servidor un flujo 
                        // tipo string UTF-8 (Independiente de la maquina)
                    salida.flush(); //  Fuerza el envio de los datos por el flujo
                                    //  de salida
		}
		catch(IOException e){;}
		txtEnvio.setText("");   //  Limpiamos textField
            }
	}   //  if(conectado)
		
	if(evt.getActionCommand().equals("Conectar..."))
            new ConectarDlg(this);  //  Presenta ventana de dialogo para esta-
                        //  blecer conexion con nuevo usuario 
    }   //  public void actionPerformed
	
    public void run(){  // Arranca el cliente, objeto ya construido
	try{
            while(true){
		if(mensaje.startsWith("/")){    //  Inicia con '/', (control) 
                    char tipo = mensaje.charAt(1); // obtengo tipo
                    mensaje=mensaje.substring(2);   //  recorto 2 primeras pos.
                    if(mensaje.matches(nick)){ //  si mensaje contiene apodo
			mensaje=">> "+mensaje;  // antepongo ">> " 
                    }
                    if(tipo=='c'){      // 'c' por conectar
			lm.addElement(mensaje);//   Incorporo usuario a la lista
                    }
                    if(tipo=='d'){      // 'd' por desconectar
                        lm.remove(lm.lastIndexOf(mensaje));//   lo excluyo
                    }
                    if(lm.getSize()==1)
			lblUsuarios.setText("1 usuario conectado");
                    else
			lblUsuarios.setText(lm.getSize() + " usuarios conectados");
                    }
		else {   //  mensaje no de control (charla)
                    txtCharla.append(mensaje + "\n");
		}
		mensaje = new String(entrada.readUTF());// leemos siguiente
            }   //  while(true)
	}
	catch (SocketException e){
            conectado=!desconectar();}
	catch (IOException e) {;}
    }   //  public void run()
	
    public boolean conectar(String ipserv, int puerto, String nick){
	try{
            lblUsuarios.setText("Conectando...");
            servidor = new Socket(ipserv, puerto);
            entrada = new DataInputStream(new BufferedInputStream(servidor.getInputStream()));
            salida = new DataOutputStream(new BufferedOutputStream(servidor.getOutputStream()));
            lblUsuarios.setText("Registrandose...");
            do{
                nick.trim();
		salida.writeUTF(nick);
		salida.flush();
		mensaje = new String(entrada.readUTF());
		if(mensaje.matches("/n")){
                    nick=JOptionPane.showInputDialog(this, 
                    "El apodo (nickname) elegido esta siendo usado por otro participante.\nElija otro por favor:","Apodo usado", JOptionPane.INFORMATION_MESSAGE);
		}
            }while(mensaje.matches("/n"));
	}   //  try
	catch(UnknownHostException e){
            JOptionPane.showMessageDialog(this, "No se pudo encontrar el servidor " + ipserv + ":" + puerto, "Error de conexin", JOptionPane.ERROR_MESSAGE);
            desconectar();
            return false;
	}
	catch(IOException e){
            JOptionPane.showMessageDialog(this, "No se pudo obtener E/S de la conexin", "Error de E/S", JOptionPane.ERROR_MESSAGE);
            desconectar();
            return false;
        }
	this.nick=nick;
	lblUsuarios.setText("Conectado.");
	receptor = new Thread(this);    //  Le abrimos un ServidorHilo  
	receptor.start();               //  Lo arrancamos
	this.setTitle("RS Chat - " +nick+" - Conectado a "+ipserv+":"+puerto);
	btnConectar.setIcon(new ImageIcon("desconectar.gif"));
	btnConectar.setLabel("Desconectar");
	lblUsuarios.setEnabled(true);
	btnEnviar.setEnabled(true);
	return true;
    }   //  public boolean conectar

    
    /* La clase interna ConectarDlg (Dialogo de conexion) que vemos a continuacion 
     * se ocupa de pedir los datos(localhost, puerto, apodo) de nuevo usuario que 
     * va a usar la ventana cliente que acabamos de arrancar. Esa ventana la 
     * podemos arrancar desde el IDE. Por ejemplo, si estamos usando NetBeans IDE, 
     * clickear Run, Run file, Run "Cliente.java" y ya en la ventana, boton "Conectar".
     * Esta clase tiene tambien comportamiento para desconectar y cerrar ventana */
    
class ConectarDlg extends JDialog implements ActionListener{
    private JLabel lblServ;
    private JLabel lblPto;
    private JLabel lblNick;
    private JTextField txtServ;
    private JTextField txtPto;
    private JTextField txtNick;
    private JButton btnConec;
    private JButton btnCancelar;
    private JDialog dlgConectar;
    
    public ConectarDlg(JFrame origen){
	super(origen,"Conectar", true);
	lblServ = new JLabel("Servidor:");
	lblPto = new JLabel("Puerto:");
	lblNick = new JLabel("Apodo (nickname):");
	txtServ = new JTextField();
	txtPto = new JTextField();
	txtNick = new JTextField();
	btnConec = new JButton("Conectar");
	btnCancelar = new JButton("Cancelar");
	lblServ.setBounds(5,5,120,25);
	lblPto.setBounds(5,35,120,25);
	lblNick.setBounds(5,65,120,25);
	txtServ.setBounds(130,5,120,25);
	txtPto.setBounds(130,35,120,25);
	txtNick.setBounds(130,65,120,25);
	btnConec.setBounds(15,95,100,25);
	btnCancelar.setBounds(15,95,100,25);
	btnConec.addActionListener(this);
	btnCancelar.addActionListener(this);
	txtServ.addActionListener(this);
	txtPto.addActionListener(this);
	txtNick.addActionListener(this);
	setSize(260,150);
	setResizable(false);
	setLocationRelativeTo(origen);
	getContentPane().setLayout(null);
	getContentPane().add(lblServ);
	getContentPane().add(lblPto);
	getContentPane().add(lblNick);
	getContentPane().add(txtServ);
	getContentPane().add(txtPto);
	getContentPane().add(txtNick);
	getContentPane().add(btnConec);
	getContentPane().add(btnCancelar);
	show();
    }   //  public ConectarDlg
		
    public void actionPerformed(ActionEvent evt){
	if(evt.getActionCommand().equals("Conectar")){
            if(txtServ.getText().length()==0 || txtPto.getText().length()==0 || txtNick.getText().length()==0){
		JOptionPane.showMessageDialog(this,"Debe ingresar todos los datos.");
            }
            else{
                int puerto;
                try{
                    puerto = Integer.parseInt(txtPto.getText());
                }
                catch(NumberFormatException e){
                    JOptionPane.showMessageDialog(this, "El puerto debe ser un nmero entero.");
                    return;
                }
                conectado=conectar(txtServ.getText(),puerto,txtNick.getText());
                dispose();
            }   //  if(txtServ.getText
	}   
	else if(evt.getActionCommand().equals("Cancelar"))dispose();
            else{
		if(txtServ.isFocusOwner())
                    txtPto.requestFocus();
		else if(txtPto.isFocusOwner())
                    txtNick.requestFocus();
		else
                    btnConec.requestFocus();
            }
        }   //  if(evt.getActionCommand()   
    }   //  public void actionPerformed

    public boolean desconectar(){
	try{
            servidor.close();
            entrada=null;
            salida=null;
            servidor=null;
            receptor=null;
            lm.clear();
            this.setTitle("RS Chat");
            btnConectar.setIcon(new ImageIcon("conectar.gif"));
            btnConectar.setLabel("Conectar...");
            lblUsuarios.setEnabled(false);
            lblUsuarios.setText("Desconectado");
            btnEnviar.setEnabled(false);
            return true;
        }
	catch (Exception ex){return false;}
    }   //  public boolean desconectar()
	
    class CierroVentana extends WindowAdapter{
	public void windowClosing(WindowEvent e){   
            btnCerrar.doClick();
        }
    }   //   class CierroVentana

    public static void main(String []args){
	Cliente ventana = new Cliente();
        ventana.setSize(610,463);
	ventana.setResizable(false);
	ventana.setTitle("RS Chat");
	ventana.setVisible(true);
    }
}   //  public boolean desconectar()
